/* sampling.c : for xlevel 
 * programmed by K. Takebe, 1997/11/02
 * Referenced:  Linux Multimedia Guide (Jeff Tranter; O'Reilly)
 */
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <fcntl.h>

#if   defined(linux)
#include <sys/soundcard.h>
#elif defined(__FreeBSD__)
#include <machine/soundcard.h>
#endif

#include "xlevel.h"

extern unsigned char buffer[NDAT_MAX]; /* in sound.c: buffer for sound data */

int        *trans_spec_list;

static int *trans_table;  /* level translation table */
static int *trans_spec;   /* level translation table */
static int level_counter = 0;
static int spec_counter  = 0;
static int dnum;

int level_L,   level_R;
int level_Lav, level_Rav;

double *lr_buf, *rr_buf, *li_buf, *ri_buf; /* buffer pointers for FFT */
double *wgt;



void init_level_trans_table()
{
  int    i;

  trans_table     = (int*)calloc(nInd, sizeof(int));

  for ( i = 0; i < nInd; i++ )
    trans_table[i] = (int)(128.0*pow(10.0,(double)(i-nInd)/(double)nInd*2.0));

  if ( trans_table == NULL ) {
    perror("");
    exit(1);
  }

  if (debug) {
    printf("\nlevel translation table \n");
    for ( i = 0; i < nInd; i++ )
      printf("\tlevel %3d -> %2d-th\n", trans_table[i], i);
  }
}


int  init_spectrum_buffer()
{
  int i;

  dnum = init_fft_param(nDat);
  if (verbose||debug)
    printf("\tFFT buffer size    = %d\n", dnum);

  lr_buf          = (double*)calloc(dnum, sizeof(double));
  rr_buf          = (double*)calloc(dnum, sizeof(double));
  li_buf          = (double*)calloc(dnum, sizeof(double));
  ri_buf          = (double*)calloc(dnum, sizeof(double));

  wgt             = (double*)calloc(dnum, sizeof(double));

  trans_spec      = (int*)calloc(nInd,    sizeof(int));
  trans_spec_list = (int*)calloc(nSpec+1, sizeof(int));

  if (lr_buf == NULL || li_buf == NULL || rr_buf == NULL || ri_buf == NULL ||
      wgt    == NULL || trans_spec == NULL || trans_spec_list == NULL) {
    perror("");
    exit(1);
  }
    
  for ( i = 0; i < dnum; i++ ) 
    wgt[i] = (0.54 + 0.46*cos((double)(i-dnum/2)*M_PI/dnum))/0.54;

  for ( i = 0; i < nInd; i++ )
    trans_spec[i] = pow(10.0, (double)(i-nInd)/(double)nInd*2.0+3.0);

  for ( i = 0; i <= nSpec; i++ )
    trans_spec_list[i] = (int)((double)dnum/2.0
			       *pow(10.0, (i-nSpec)/(double)nSpec*2.0));

  for ( i = 0; i <= nSpec; i++ ) {
    if ( trans_spec_list[i] == 0 )
      trans_spec_list[i] = 1; /* ignore direct current component */
  }
  if ( trans_spec_list[nSpec] >= dnum/2 )
    trans_spec_list[nSpec] = dnum/2-1;

  if (debug) {
    printf("\nferq-level translation table \n");
    for ( i = 0; i < nInd; i++ )
      printf("\tlevel %3d -> %2d-th\n", trans_spec[i], i);

    printf("\nfreq-band table \n");
    for ( i = 0; i < nSpec; i++ ) {
      printf("\tband %2d : %5d Hz - %5d Hz", i,
	     (int)(sampling*(double)trans_spec_list[i]/(double)dnum),
	     (int)(sampling*(double)trans_spec_list[i+1]/(double)dnum));
      if (debug)
	printf(" : %d - %d", trans_spec_list[i], trans_spec_list[i+1]);

      printf("\n");
    }
  }
  return 0;
}


void store_level_data()
{
  register unsigned char *p;
  unsigned char *endp;

  get_data();

  if ( level_counter == 0 ) {
    level_L = level_Lav = 0;
    level_R = level_Rav = 0;
  }

  endp = buffer+nDat*2;
  for ( p = buffer; p < endp; ) {
    level_Lav = level_Lav + (*p - 127);
    level_L   = level_L   + abs(*(p++) - 127);
    level_Rav = level_Rav + (*p - 127);
    level_R   = level_R   + abs(*(p++) - 127);
  }

  level_counter++;
}


void get_level(int *left, int *right)
{
  register int  i;
  int           l, r;   

  if ( level_counter == 0 )
    store_level_data();

  l = (level_L - abs(level_Lav))/(nDat*level_counter);
  r = (level_R - abs(level_Rav))/(nDat*level_counter);

  for ( i = 0; i < nInd-1 && l > trans_table[i]; i++ );
  *left  = i;

  for ( i = 0; i < nInd-1 && r > trans_table[i]; i++ ); 
  *right = i;

  level_counter = 0;
}


void store_spectrum_data()
{
  register int  i;
  register unsigned char *p;

  get_data();

  if (spec_counter == 0) {
    for ( p = buffer, i = 0; i < dnum; i++) {
      lr_buf[i] = (double)(*(p++) - 127);
      rr_buf[i] = (double)(*(p++) - 127);
    }
  } else {
    for ( p = buffer, i = 0; i < dnum; i++) {
      lr_buf[i] = lr_buf[i] + (double)(*(p++) - 127);
      rr_buf[i] = rr_buf[i] + (double)(*(p++) - 127);
    }
  }
  spec_counter++;
}


void get_spectrum(int *left, int *right)
{
  register int    i, j;
  static double d = 1.0/128.0;
  int      nDat2 = dnum/2;
  double   w;
  double   l, r;   
  double   l_rp, l_ip;
  double   r_rp, r_ip;

  if (spec_counter == 0) store_spectrum_data();

  for ( i = 0; i < dnum; i++) {
    w = wgt[i];
    lr_buf[i] = lr_buf[i]*d*w;
    rr_buf[i] = rr_buf[i]*d*w;
    li_buf[i] = 0.0;
    ri_buf[i] = 0.0;
  }

  fft1(lr_buf, li_buf);
  fft1(rr_buf, ri_buf);

  for ( j = 0; j < nSpec; j++ ) {
    l = r = 0.0;
    if ( trans_spec_list[j] == trans_spec_list[j+1] ) {
      i = trans_spec_list[j];
      l_rp = lr_buf[i]; l_ip = li_buf[i];
      r_rp = rr_buf[i]; r_ip = ri_buf[i];

      l = sqrt(l_rp*l_rp + l_ip*l_ip);
      r = sqrt(r_rp*r_rp + r_ip*r_ip);
    } else
      for (i = trans_spec_list[j]; i < trans_spec_list[j+1] && i<nDat2; i++) {
	l_rp = lr_buf[i]; l_ip = li_buf[i];
	r_rp = rr_buf[i]; r_ip = ri_buf[i];
	
	l = l + sqrt(l_rp*l_rp + l_ip*l_ip);
	r = r + sqrt(r_rp*r_rp + r_ip*r_ip);
      }
    for ( i = 0; i < nInd-1 && l > trans_spec[i]; i++ );
    left[j]  = i;
    for ( i = 0; i < nInd-1 && r > trans_spec[i]; i++ ); 
    right[j] = i;
  }
  spec_counter = 0;
}
